#!/usr/bin/env python3
"""
Windows Port Scanner (Network Security Tool)

Parameters:
    --target IP: Target IP address
    --ports RANGE: Port range to scan
    --type TYPE: Scan type (TCP/UDP)
    --timeout SEC: Connection timeout
    --threads N: Number of threads
    --format FORMAT: Output format (table/json)

Requires: socket, concurrent.futures, tabulate packages
Warning: Only scan networks you have permission to test
"""

import socket
import sys
import threading
import time
import json
import argparse
import concurrent.futures
from typing import List, Dict, Any, Optional, Tuple
from tabulate import tabulate
import logging

class PortScanner:
    def __init__(self, timeout: float = 1.0, max_threads: int = 100):
        self.timeout = timeout
        self.max_threads = max_threads
        self.logger = logging.getLogger(__name__)
        self.common_ports = {
            20: "FTP-DATA", 21: "FTP", 22: "SSH", 23: "TELNET",
            25: "SMTP", 53: "DNS", 80: "HTTP", 110: "POP3",
            143: "IMAP", 443: "HTTPS", 445: "SMB", 3389: "RDP"
        }

    def scan_port(self, ip: str, port: int) -> Dict[str, Any]:
        """Scan a single port."""
        result = {
            'port': port,
            'state': 'closed',
            'service': self.common_ports.get(port, 'unknown'),
            'banner': None
        }
        
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(self.timeout)
                if sock.connect_ex((ip, port)) == 0:
                    result['state'] = 'open'
                    try:
                        sock.send(b'')
                        banner = sock.recv(1024).decode().strip()
                        if banner:
                            result['banner'] = banner
                    except:
                        pass
                        
        except Exception as e:
            self.logger.debug(f"Error scanning port {port}: {str(e)}")
            
        return result

    def scan_ports(self, ip: str, start_port: int = 1,
                  end_port: int = 65535) -> List[Dict[str, Any]]:
        """Scan a range of ports using multiple threads."""
        results = []
        ports = range(start_port, end_port + 1)
        
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_threads) as executor:
            future_to_port = {
                executor.submit(self.scan_port, ip, port): port
                for port in ports
            }
            
            for future in concurrent.futures.as_completed(future_to_port):
                try:
                    result = future.result()
                    if result['state'] == 'open':
                        results.append(result)
                except Exception as e:
                    port = future_to_port[future]
                    self.logger.error(f"Error scanning port {port}: {str(e)}")
                    
        return sorted(results, key=lambda x: x['port'])

    def display_results(self, results: List[Dict[str, Any]], 
                       output_format: str = 'table'):
        """Display scan results in the specified format."""
        if not results:
            print("No open ports found.")
            return
            
        if output_format == 'json':
            print(json.dumps(results, indent=4))
        else:
            headers = ['Port', 'State', 'Service', 'Banner']
            table_data = [
                [r['port'], r['state'], r['service'], 
                 r['banner'] if r['banner'] else ''] 
                for r in results
            ]
            print(tabulate(table_data, headers=headers, tablefmt='grid'))

def main():
    parser = argparse.ArgumentParser(
        description='Port Scanner - Scan ports on a target host')
    parser.add_argument('ip', help='Target IP address')
    parser.add_argument('--start', type=int, default=1,
                       help='Start port (default: 1)')
    parser.add_argument('--end', type=int, default=65535,
                       help='End port (default: 65535)')
    parser.add_argument('--timeout', type=float, default=1.0,
                       help='Timeout in seconds (default: 1.0)')
    parser.add_argument('--threads', type=int, default=100,
                       help='Maximum number of threads (default: 100)')
    parser.add_argument('--format', choices=['table', 'json'],
                       default='table', help='Output format')
    
    args = parser.parse_args()
    
    # Setup logging
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )
    
    try:
        scanner = PortScanner(timeout=args.timeout, max_threads=args.threads)
        
        print(f"Scanning ports {args.start}-{args.end} on {args.ip}...")
        start_time = time.time()
        
        results = scanner.scan_ports(args.ip, args.start, args.end)
        
        scan_time = time.time() - start_time
        print(f"\nScan completed in {scan_time:.2f} seconds")
        
        scanner.display_results(results, args.format)
        
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}")
        return 1
        
    return 0

if __name__ == "__main__":
    exit(main())
